# pytest functions for decorators

# Generated by CodiumAI

from wizard._utils.decorators import check_path
from wizard._utils.decorators import track_execution_time
from wizard._utils.decorators import check_limits
from wizard._utils.decorators import add_method


import time
import pytest
import numpy as np

VALID_PATH = '.'


class TestCheckPath:

    #  Function called with valid path
    def test_valid_path(self):
        @check_path
        def dummy_func(path):
            return True

        assert dummy_func(VALID_PATH) == True

    #  Function called with valid path and additional arguments
    def test_valid_path_with_args(self):
        @check_path
        def dummy_func(path, arg1, arg2):
            return True

        assert dummy_func(VALID_PATH, 'arg1', 'arg2') == True

    #  Function called with valid path and keyword argument
    def test_valid_path_with_kwarg(self):
        @check_path
        def dummy_func(path, kwarg=None):
            return True

        assert dummy_func(VALID_PATH, kwarg='value') == True

    #  Function called with valid path and multiple keyword arguments
    def test_valid_path_with_multiple_kwargs(self):
        @check_path
        def dummy_func(path, kwarg1=None, kwarg2=None):
            return True

        assert dummy_func(VALID_PATH, kwarg1='value1', kwarg2='value2') == True

    #  Function called with empty string path
    def test_empty_string_path(self):
        @check_path
        def dummy_func(path):
            return True

        with pytest.raises(ValueError):
            dummy_func('')

    #  Function called with non-existent path
    def test_nonexistent_path(self):
        @check_path
        def dummy_func(path):
            return True

        with pytest.raises(FileNotFoundError):
            dummy_func('/path/to/nonexistent')

class TestCheckLimits:

    #  The function receives an image with all values equal to the upper limit and returns the same image.
    def test_all_values_equal_to_upper_limit_returns_same_image(self):
        # Arrange
        image = np.array([[1, 1], [1, 1]], dtype='float32')

        @check_limits
        def dummy_func(image):
            return image

        # Act
        result = dummy_func(image)

        # Assert
        assert np.array_equal(result, image)

    #  The function receives an image with all values equal to the lower limit and returns the same image.
    def test_all_values_equal_to_lower_limit_returns_same_image(self):
        # Arrange
        image = np.array([[0, 0], [0, 0]], dtype='float32')

        @check_limits
        def dummy_func(image):
            return image

        # Act
        result = dummy_func(image)

        # Assert
        assert np.array_equal(result, image)

    #  The function receives an empty image and returns an empty image.
    def test_empty_image_returns_empty_image(self):
        # Arrange
        image = np.array([], dtype='float32')

        @check_limits
        def dummy_func(image):
            return image

        # Act
        result = dummy_func(image)

        # Assert
        assert np.array_equal(result, image)

    #  The function receives an image with a single value and returns the same image.
    def test_single_value_image_returns_same_image(self):
        # Arrange
        image = np.array([100], dtype='float32')

        @check_limits
        def dummy_func(image):
            return image

        # Act
        result = dummy_func(image)

        # Assert
        assert np.array_equal(result, [1.])

    #  The function receives an image with negative values and returns the clipped image.
    def test_negative_values_returns_clipped_image(self):
        # Arrange
        image = np.array([[-100, -200], [150, 250]], dtype='float32')

        @check_limits
        def dummy_func(image):
            return image

        # Act
        result = dummy_func(image)

        # Assert
        assert np.array_equal(result, np.array([[0., 0.], [1., 1.]], dtype='float32'))

    #  The function receives an image with values above the upper limit and returns the clipped image.
    def test_values_above_upper_limit_returns_clipped_image(self):
        # Arrange
        image = np.array([[300, 400], [500, 600]], dtype='float32')

        @check_limits
        def dummy_func(image):
            return image

        # Act
        result = dummy_func(image)

        # Assert
        assert np.array_equal(result, np.array([[1, 1], [1, 1]], dtype='float32'))


class TestAddMethod:

    #  The decorator function should return a function.
    def test_decorator_returns_function(self):

        class MyClass:
            pass

        @add_method(MyClass)
        def my_method():
            return None

        assert callable(MyClass.my_method)



# Configuration file for the Sphinx documentation builder.
#
# For the full list of built-in configuration values, see the documentation:
# https://www.sphinx-doc.org/en/master/usage/configuration.html

# -- Project information -----------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#project-information

project = 'hsi_wizard'
copyright = '2024, Felix Wühler'
author = 'Felix Wühler'
release = '2024-01-18'

# -- General configuration ---------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#general-configuration

extensions = [
    'sphinx.ext.autodoc',           # To generate autodocs
    'sphinx.ext.mathjax',           # autodoc with maths
    'sphinx.ext.napoleon'           # For auto-doc configuration
]

napoleon_google_docstring = False   # Turn off googledoc strings
napoleon_numpy_docstring = True     # Turn on numpydoc strings
napoleon_use_ivar = True            # For maths symbology

templates_path = ['_templates']
exclude_patterns = ['_build', 'Thumbs.db', '.DS_Store']



# -- Options for HTML output -------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#options-for-html-output

html_theme = 'alabaster'
html_static_path = ['_static']



"""
This code was inspired by the repository:

Repository: specio
Author/Organization: paris-saclay-cds
Original repository: https://github.com/paris-saclay-cds/specio
"""
import struct
import numpy as np


def _block_info(data):
    """Retrieve the information of the next block."""
    if len(data) != 6:
        raise ValueError(f"'data' should be 6 bytes. Got {len(data)} instead.")
    return struct.unpack('<Hi', data)


def _decode_5100(data):
    """Read the block of data with ID 5100."""
    name_size = struct.unpack('<h', data[:2])[0]
    name = data[2:name_size + 2].decode('utf8')
    header_format = '<ddddddddddiiihBhBhBhB'
    header_data = struct.unpack(header_format, data[name_size + 2:name_size + 106])
    return {'name': name, 'x_delta': header_data[0], 'y_delta': header_data[1], 'z_delta': header_data[2], 
            'z_start': header_data[3], 'z_end': header_data[4], 'z_4d_start': header_data[5], 'z_4d_end': header_data[6],
            'x_init': header_data[7], 'y_init': header_data[8], 'z_init': header_data[9], 'n_x': header_data[10], 
            'n_y': header_data[11], 'n_z': header_data[12], 'resolution': header_data[17], 'transmission': header_data[19]}


def _decode_5104(data):
    """Read the block of data with ID 5104."""
    text = []
    start_byte = 0
    data_length = len(data)

    while start_byte + 2 <= data_length:
        tag = data[start_byte:start_byte + 2]
        start_byte += 2
        
        if tag == b'#u':
            text_size = struct.unpack('<h', data[start_byte:start_byte + 2])[0]
            start_byte += 2
            text.append(data[start_byte:start_byte + text_size].decode('utf8'))
            start_byte += text_size + 6  # Skip over text and padding
        elif tag == b'$u' or tag == b',u':
            text_size = struct.unpack('<h', data[start_byte:start_byte + 2])[0]
            start_byte += 2
            text.append(text_size)  # Here we assume it's an integer, as in the original code
            start_byte += 6 if tag == b'$u' else 0
        else:
            start_byte -= 1  # Step back if the tag is not recognized, to avoid getting stuck

    # Create the dictionary to return
    keys = ['analyst', 'date', 'image_name', 'instrument_model', 
            'instrument_serial_number', 'instrument_software_version', 
            'accumulations', 'detector', 'source', 'beam_splitter', 
            'apodization', 'spectrum_type', 'beam_type', 'phase_correction', 
            'ir_accessory', 'igram_type', 'scan_direction', 'background_scans', 
            'ir_laser_wave_number_unit']
    
    # Ensure we don't run into an IndexError if the text list is too short
    text_dict = {key: (text[i] if i < len(text) else None) for i, key in enumerate(keys)}

    return text_dict


def _decode_5105(data):
    """Read the block of data with ID 5105."""
    return np.frombuffer(data, dtype=np.float32)


FUNC_DECODE = {5100: _decode_5100, 5104: _decode_5104, 5105: _decode_5105}


def _read_fsm(fsm_file):
    """Read the FSM file and extract spectrum data."""
    content = open(fsm_file, "rb").read()
    start_byte = 44
    meta = {'signature': content[:4], 'description': content[4:44].decode('utf-8'), 'filename': fsm_file}
    spectrum = []

    while start_byte < len(content):
        block_id, block_size = _block_info(content[start_byte:start_byte + 6])
        start_byte += 6
        if block_size == 0:
            continue
        block_data = content[start_byte:start_byte + block_size]
        data_extracted = FUNC_DECODE[block_id](block_data)
        start_byte += block_size
        if isinstance(data_extracted, dict):
            meta.update(data_extracted)
        else:
            spectrum.append(data_extracted)

    wavelength = np.arange(meta['z_start'], meta['z_end'] + meta['z_delta'], meta['z_delta'])
    return np.squeeze(spectrum), wavelength, meta



def find_nex_greater_wave(waves, wave_1: int, maximum_deviation: int = 5):
    """
    Finds the next greater wave value in a list of waves within a specified deviation.

    Given a list of wave values, this function identifies the smallest wave value greater than the specified wave_1 within the range defined by maximum_deviation. If no such value exists within the range, it returns -1.

    :param waves: A list of integers representing the available wave values.
    :type waves: list[int]
    :param wave_1: The starting wave value to find the next greater wave for.
    :type wave_1: int
    :param maximum_deviation: The maximum deviation from wave_1 to consider for finding the next greater wave.
    :type maximum_deviation: int
    :returns: The next greater wave value within the deviation range, or -1 if no such value exists.
    :rtype: int
    """

    wave_next = -1

    for n in range(maximum_deviation):
        wave_n = wave_1 + n

        if wave_n in waves:
            wave_next = wave_n
            break

    return wave_next


def find_nex_smaller_wave(waves, wave_1: int, maximum_deviation: int = 5):
    """
    Finds the next smaller wave value in a list of waves within a specified deviation.
    
    Given a list of wave values, this function identifies the largest wave value smaller than the specified wave_1 within the range defined by maximum_deviation. If no such value exists within the range, it returns -1.
    
    :param waves: A list of integers representing the available wave values.
    :type waves: list[int]
    :param wave_1: The starting wave value to find the next smaller wave for.
    :type wave_1: int
    :param maximum_deviation: The maximum deviation from wave_1 to consider for finding the next smaller wave.
    :type maximum_deviation: int
    :returns: The next smaller wave value within the deviation range, or -1 if no such value exists.
    :rtype: int
    """

    wave_next = -1

    for n in range(maximum_deviation):
        wave_n = wave_1 - n

        if wave_n in waves:
            wave_next = wave_n
            break

    return wave_next



import copy
import numpy as np
import matplotlib.pyplot as plt
from wizard import DataCube
from .._utils.helper_functions import find_nex_smaller_wave

try:
    from IPython import get_ipython
    if "IPKernelApp" not in get_ipython().config:
        raise ImportError("console")
except ImportError:
    import matplotlib
    matplotlib.use('TkAgg')
except Exception as e:
    # print(f"An unexpected error occurred: {e}")
    pass


def normalize_layer(layer: np.array) -> np.array:
    """
    Normalize a 2D or 3D numpy array (layer) by adjusting its values to a 0-1 range.

    This function first checks if the maximum value in the layer is more than 10 times greater than the mean value and prints a warning if so. It then normalizes the layer by removing any offset and scaling the values to the range [0, 1]. The resulting layer is rounded and cast to `float16` for consistency.

    :param layer: A numpy array representing the layer to be normalized.
    :type layer: np.ndarray
    :returns: The normalized layer as a numpy array with values scaled to the range [0, 1] and rounded to 10 decimal places.
    :rtype: np.ndarray
    :raises ValueError: If the input `layer` is not a numpy array.
    """
    if layer.max() > layer.mean() * 10:
        print('\033[93mThe layer max value is more than 10 times greater than the mean value.'
              ' If you don’t see anything, try the spike removing tool.\033[0m')
    
    layer_copy = copy.deepcopy(layer)

    if layer_copy.min() != 0:
        layer_copy -= layer.min()

    if layer_copy.max() != 0:
        layer_copy /= layer_copy.max()

    layer_copy = np.round(layer_copy, decimals=10).astype('float16')

    return layer_copy


def plotter(dc: DataCube) -> None:
    """
    Interactive plotter for visualizing and analyzing data from a DataCube.

    This function creates an interactive plot with two subplots: one for displaying an image layer from the DataCube and another for displaying a spectral plot. Users can interact with the plots to select different image layers and regions of interest, which updates both the image and spectral plot accordingly.

    :param dc: A DataCube object containing the data to be visualized. The DataCube should have attributes `wavelengths` and `cube`, where `cube` is a 3D numpy array and `wavelengths` is a list or array of wavelength values corresponding to the layers of the cube.
    :type dc: DataCube
    :returns: None
    :raises AttributeError: If the DataCube object does not have the required attributes (`wavelengths` and `cube`).
    :raises ValueError: If the DataCube object does not have a `name` attribute, but the code assumes it exists.

    Example usage:
    >>> plotter(my_datacube)
    """
    global layer_id
    global x_id
    global y_id

    layer_id = dc.wavelengths[0]
    x_id = 0
    y_id = 0

    def update_plot(val):
        layer = dc.cube[np.where(dc.wavelengths == layer_id)[0][0]]
        layer = normalize_layer(layer)
        imshow.set_data(layer)

        spec = dc.cube[:, x_id, y_id]
        wave = range(dc.cube.shape[0]) if dc.wavelengths is None else dc.wavelengths
        s_plot.set_data(wave, spec)
        r = (spec.max() - spec.min()) * 0.1
        ax[1].set_ylim(spec.min() - r, spec.max() + r)
        line.set_data([layer_id], (layer.min(), layer.max()))
        fig.canvas.draw_idle()

    def onclick_select(event):
        global layer_id
        global x_id
        global y_id

        if event.inaxes == ax[0]:
            y_id = int(event.xdata)
            x_id = int(event.ydata)
            update_plot(event)

        elif event.inaxes == ax[1]:
            tmp_id = int(event.xdata)
            layer_id = find_nex_smaller_wave(dc.wavelengths, tmp_id, 10)
            if layer_id != -1:
                update_plot(event)

    fig, ax = plt.subplots(1, 2, figsize=(12, 6))
    plt.subplots_adjust(bottom=0.25)
    fig.suptitle('Datacube' if dc.name is None else dc.name)

    layer = normalize_layer(dc.cube[0])
    imshow = ax[0].imshow(layer)

    spec = dc.cube[:, 0, 0]
    wave = range(dc.cube.shape[0]) if dc.wavelengths is None else dc.wavelengths

    line = ax[1].axvline(
        x=layer_id,
        color='lightgrey',
        linestyle='dashed',
    )

    s_plot, = ax[1].plot(wave, spec)

    fig.canvas.mpl_connect("button_press_event", onclick_select)

    update_plot(None)
    plt.show()



# wizard/__init__.py


# Import necessary submodules and classes/functions from them
from ._core.datacube import DataCube
from ._core.eda import plotter

from ._utils.data_loader import read
# from ._utils.utils_module1 import UtilityClass1, util_function1
# from ._utils.utils_module2 import UtilityClass2, util_function2

# Define what should be accessible when using 'from wizard import *'
__all__ = [
    'DataCube',
    'read'
]

# Example of setting package metadata
__version__ = "0.0.1"
__author__ = 'flx'


# Optionally, define a function for package initialization or configuration
def initialize():
    pass


# Initialize the package
initialize()



"""Decorators to handle behavier."""
import os
import time

from functools import wraps

import numpy as np


def check_load_dc(func) -> np.array:
    """
    Check if the loading function is correctly defined.

    :param func:
    :return: func
    :rtype: method
    """
    def wrapper(*args, **kwargs):
        # process function
        cube = func(*args, **kwargs)

        if cube != 'no implementation':
            # check if np array
            if cube is not np.array:
                raise ValueError('Loading function should return a np array')

            if 2 < len(cube.shape) <= 3:
                raise ValueError('loading function is not valid. The return'
                                 'shape should be (v|x|y)')
        else:
            cube = None
        return cube
    return wrapper


def check_path(func):
    """
    Check if data path is valid, if not throw exception.

    :param func:
    :return:
    """
    def wrapper(*args, **kwargs):

        path = args[0] if kwargs.get('path') is None else kwargs.get('path')

        if path is None or path == '':
            raise ValueError('no path given')

        if not os.path.exists(path):
            raise FileNotFoundError(f'Given path is not valid. {path}')

        # process function
        return func(*args, **kwargs)

    return wrapper


def add_method(cls):
    """
    Add method to class.

    :param cls:
    :return:

    source: [Michael Garod @ Medium](https://mgarod.medium.com/
    dynamically-add-a-method-to-a-class-in-python-c49204b85bd6)
    """
    def decorator(func):
        @wraps(func)
        # def wrapper(self, *args, **kwargs):
        def wrapper(*args, **kwargs):
            return func(*args, **kwargs)
        setattr(cls, func.__name__, wrapper)
        # Note we are not binding func, but wrapper which accepts self but
        # does exactly the same as func
        return func  # returning func means func can still be used normally
    return decorator


def track_execution_time(func):
    """
    Decorator function to track the execution time of a function in milliseconds.

    Parameters
    ----------
    func : callable
        The function to be decorated.

    Returns
    -------
    callable
        A wrapped function that executes `func` and prints the execution time in milliseconds.

    Notes
    -----
    This decorator uses `time.time()` to measure the start and end times of `func` execution,
    computes the elapsed time, and prints it formatted to four decimal places.

    Example
    -------
    >>> @track_execution_time
    ... def example_function():
    ...     return sum(range(1000000))
    ...
    >>> example_function()
    Function 'example_function' executed in 2.1200 ms
    """
    @wraps(func)
    def wrapper(*args, **kwargs):
        start_time = time.time()
        result = func(*args, **kwargs)
        end_time = time.time()
        execution_time = (end_time - start_time) * 1000
        print(f"Function '{func.__name__}' executed in {execution_time:.4f} ms")
        return result
    return wrapper


def add_to_workflow(func):
    """
    Add a function and the parameter to an template.

    :param func:
    :return:
    """
    def wrapper(*args, **kwargs):
        print(*args, **kwargs)

        return func(*args, **kwargs)
    return wrapper()


def check_limits(func) -> np.array:
    """
    Force clipping limits to an image or array.

    :param func:
    :return:
    """
    def wrapper(*args, **kwargs):
        # store data type
        dtype = args[0].dtype

        # process function
        image = func(*args, **kwargs)

        if dtype in ['float32', 'float64']:
            image = np.clip(image, 0, 1).astype(dtype)

        return image
    return wrapper



import numpy as np
from .._utils.decorators import check_limits


def extend_image(img: np.array, extend_x: int, extend_y: int) -> np.array:
    """
    Extend image by 2 * extend_x and 2 * extend_y.

    Extend_x/y adds zero borders in the given size.

    :param img: The input image.
    :param extend_x: The extension size along the x-axis.
    :param extend_y: The extension size along the y-axis.
    :return: Extended image.
    :rtype: np.array
    """
    x, y, z = img.shape

    new_img = np.zeros(
        shape=(x + extend_x * 2, y + extend_y * 2, z),
        dtype=img.dtype
    )
    new_img[extend_x:-extend_x, extend_y:-extend_y] = img

    return new_img


def decrease_image(img: np.array, decrease_x: int, decrease_y: int) -> np.array:
    """
    Decrease image by 2 * decrease_x and 2 * decrease_y.

    Decrease_x/y removes borders from the given size.

    :param img: The input image.
    :param decrease_x: The decrease size along the x-axis.
    :param decrease_y: The decrease size along the y-axis.
    :return: Decreased image.
    :rtype: np.array
    """
    return img[decrease_x:-decrease_x, decrease_y:-decrease_y]


def get_output_size(image_lengths: int, filter_lengths: int, stride: int) -> int:
    """
    Calculate length of the feature map.

    The function should return an integer value.
    If the math is not possible, the function returns 0.

    :param image_lengths: The length of one image side.
    :type image_lengths: int
    :param filter_lengths: The length of one side of the filter.
    :type filter_lengths: int
    :param stride: Filter stride.
    :type stride: int
    :return: Feature lengths as an integer.
    :rtype: int

    >>> get_output_size(10, 2, 1)
    9

    >>> get_output_size(10, 3, 2)
    0
    """
    feature_lengths = (image_lengths - filter_lengths) / stride + 1
    return int(feature_lengths) if feature_lengths.is_integer() else 0


def feature_map(img: np.array, filter: np.array, padding: str = 'const', stride_x: int = 1, stride_y: int = 1):
    """
    Generate a feature map by stepping over the image with a given filter function.

    :param img: The input image.
    :param filter: The filter to apply.
    :param padding: Padding type.
    :param stride_x: Stride along the x-axis.
    :param stride_y: Stride along the y-axis.
    :return: Feature map.
    """
    if stride_x == 0:
        raise ValueError('step_x cant be 0 (zero)')
    elif stride_y == 0:
        raise ValueError('step_y cant be 0 (zero)')

    feature_map_len_x = get_output_size(img.shape[0], filter.shape[0], stride_x)
    feature_map_len_y = get_output_size(img.shape[1], filter.shape[1], stride_y)

    if feature_map_len_x == 0:
        while feature_map_len_x == 0:
            stride_x += 1
            feature_map_len_x = get_output_size(img.shape[0], filter.shape[0], stride_x)
        return None
    if feature_map_len_y == 0:
        while feature_map_len_y == 0:
            stride_y += 1
            feature_map_len_y = get_output_size(img.shape[1], filter.shape[1], stride_y)
        return None

    feature_img = np.zeros(
        shape=(
            feature_map_len_x,
            feature_map_len_y,
            img.shape[2]
        )
    )

    for x in range(feature_map_len_x):
        x1 = x * stride_x
        x2 = x * stride_x + filter.shape[0]
        for y in range(feature_map_len_y):
            y1 = y * stride_y
            y2 = y * stride_y + filter.shape[1]
            mini_img = img[x1:x2, y1:y2]
            pixel = np.sum(mini_img * filter)
            feature_img[x, y] = pixel

    feature_img = feature_img / feature_img.max()

    return feature_img


@check_limits
def rgb_to_grayscale_average(image: np.array) -> np.array:
    """
    Convert RGB image to grayscale using average values.

    Color/3 to avoid conflicts with uint8 images.

    :param image: The RGB image.
    :return: Grayscale image.
    """
    return image[:, :, 0] / 3 + image[:, :, 1] / 3 + image[:, :, 2] / 3


@check_limits
def rgb_to_grayscale_wight(image: np.array, r_weight: float = 0.299, g_weight: float = 0.587,
                           b_weight: float = 0.114) -> np.array:
    """
    Convert RGB image to grayscale using weighted average.

    :param image: The RGB image.
    :param r_weight: Weight for adjusting red color.
    :param g_weight: Weight for adjusting green color.
    :param b_weight: Weight for adjusting blue value.
    :return: Grayscale image.
    """
    return image[:, :, 0] * r_weight + image[:, :, 1] * g_weight + image[:, :, 2] * b_weight


@check_limits
def brightness(image: np.array, delta: int) -> np.array:
    """
    Adjust brightness of an image.

    :param image: The input image.
    :param delta: The delta value for brightness adjustment.
    :return: Brightness adjusted image.
    """
    tmp = image + delta
    tmp[tmp < delta] = 255
    return tmp


@check_limits
def contrast(image: np.array, beta: int) -> np.array:
    """
    Adjust contrast of an image.

    :param image: The input image.
    :param beta: The beta value for contrast adjustment.
    :return: Contrast adjusted image.
    """
    u = np.mean(image, axis=2)
    u_mean = u.mean()

    if beta == 255:
        alpha = np.infty
    else:
        alpha = (255 + beta) / (255 - beta)

    image = ((image[:, :] - u_mean) * alpha + u_mean).astype('int')
    return image



"""DataCube class for storring HSI data."""
import warnings

import cv2
from rich import print
import numpy as np
import pickle

from .._utils import decorators
from .._utils.tracker import TrackExecutionMeta


class DataCube(metaclass=TrackExecutionMeta):
    """
    DataCube class to store several important information for the hsi data.

    The cube is a 3D array with shape `vxy`.
    x, y values describe the `pixel`.
    v values are for the measured counts, channels, values, ...

    In the most cases you have some kind of wavelengths information in your dc.
    These can be nm or cm^-1. You can applie your description in notation for
    the dc and cast it.
    """

    def __init__(self, cube=None, wavelengths=None, name=None,
                 notation=None, record: bool = False) -> None:
        """
        Magic Methods to init a new instance.

        :param cube: dc
        :param wavelengths: wavelengths of the spectral information as list
        :param name: name of the dc
        :param notation: wavelengths or wavenumbers
        :return: None
        :rtype return: None
        """
        self.name = name  # name of the dc
        self.shape = None if cube is None else cube.shape  # shape of the dc
        self.dim = None  # get dimension of the dc 2d, 3d, 4d ...
        self.wavelengths = np.array(wavelengths) if wavelengths is not None \
            else np.arange(0, cube.shape[0], dtype=int) if cube is not None \
            else None
        self.cube = None if cube is None else cube
        self.notation = notation

        self.record = record
        if self.record:
            self.start_recording()

    def __add__(self, other):
        """
        Magic Method to add two Datacubes.

        :param other: DataCube that should be added
        :return: None
        :type: None
        """
        if not isinstance(other, DataCube):
            raise ValueError('Cant add DataCube and none DataCube.')

        new_wavelengths = None

        if self.cube.shape[1:] != other.cube.shape[1:]:
            raise ValueError(
                f'DataCubes needs to have the same `x` an `y` shape.\n'
                f'Cube1: {self.cube.shape}, Cube2: {other.cube.shape}'
                f'You can use the DataCube.resize function to adjust the cubes'
            )

        # wavelengths cant be empty atm
        if self.wavelengths is None or other.wavelengths is None:
            warnings.warn('One of the two DataCubes does not contain the'
                          ' wavelength information. Adding them will work,'
                          ' but you will lose this information.')
        else:
            new_wavelengths = self.wavelengths + other.wavelengths

        if self.cube is None or other.cube is None:
            raise ValueError("Cannot add DataCubes with None values.")
        new_cube = np.concatenate((self.cube, other.cube), axis=0)
        return DataCube(cube=new_cube, wavelengths=new_wavelengths,
                        name=self.name, notation=self.notation)

    def __len__(self) -> int:
        """
        Magic Method for getting length of axis as int.

        :param axis: int for selection axis
        :return: length of DataCube for given axis
        :rtype: int
        """
        return self.shape[0] if self.cube is not None else 0

    def __getitem__(self, idx):
        """
        Magic Method to get an item.

        :param idx:
        :return:

        """
        return self.cube[idx]

    def __setitem__(self, idx, value) -> None:
        """
        Magic Method to set an item.

        :param idx:
        :param value:
        :return: None
        """
        self.cube[idx] = value

    def __iter__(self):
        """
        Magic Method to iter ofer DataCube.

        :return:
        """
        self.idx = 0
        return self

    def __next__(self):
        """
        Magic Method for next.

        :return:
        """
        if self.idx >= len(self.cube):
            raise StopIteration
        else:
            self.idx += 1
            return self.cube[self.idx - 1], self.wavelengths[self.idx - 1]

    # def __sizeof__(self):
    #   pass

    def __str__(self) -> str:
        """
        Magic Method, print dc information.

        :return : string with dc information
        :rtype: str
        """
        n = '\n'
        _str = ''
        _str += f'Name: {self.name}' + n
        _str += f'Shape: {self.shape}' + n
        if self.wavelengths is not None:
            _str += 'Wavelengths' + n
            _str += f'Num: {len(self.wavelengths)}' + n
            _str += f'From: {self.wavelengths.min()}' + n
            _str += f'To: {self.wavelengths.max()}' + n
        _str += 'Cube:' + n
        _str += f'{self.cube}' + n
        return _str

    def execute_template(self, template_data) -> None:
        """
        Execute Template.

        :return:
        """
        for method_name, args, kwargs in template_data:
            method = getattr(self, method_name)
            method(*args, **kwargs)

    @decorators.check_load_dc
    def load(self, *args, **kwargs) -> None:
        """
        Empty load Function to override.

        This is a template function. You can implemnt your own load functions.

        :return:
        """
        raise NotImplementedError('Subclasses must implement the `load`'
                                  'method')

    def resize(self, x_new: int, y_new: int,
               interpolation: str = 'linear') -> None:
        """
        Resize DataCube.cube.

        cv2.INTER_LINEAR 	The standard bilinear interpolation, ideal for
                            enlarged images.
        cv2.INTER_NEAREST 	The nearest neighbor interpolation, which, though
                            fast to run, creates blocky images.
        cv2.INTER_AREA 	 The interpolation for the pixel area, which scales
                            down images.
        cv2.INTER_CUBIC 	The bicubic interpolation with 4×4-pixel
                            neighborhoods, which, though slow to run, generates
                            high-quality instances.
        cv2.INTER_LANCZOS4 	The Lanczos interpolation with an 8×8-pixel
                            neighborhood, which generates images of the highest
                            quality but is the slowest to run.

        :param interpolation:
        :param x_new:
        :param y_new:
        :return:
        """
        mode = None

        shape = self.cube.shape

        if shape[1] > x_new:
            print('\033[93mx_new is smaller then the exising cube,'
                  'you lose information\033[0m')
        if shape[2] > y_new:
            print('\033[93my_new is smaller then the exising cube,'
                  'you lose information\033[0m')

        if interpolation == 'linear':
            mode = cv2.INTER_LINEAR
        elif interpolation == 'nearest':
            mode = cv2.INTER_NEAREST
        elif interpolation == 'area':
            mode = cv2.INTER_AREA
        elif interpolation == 'cubic':
            mode = cv2.INTER_CUBIC
        elif interpolation == 'Lanczos':
            mode = cv2.INTER_LANCZOS4

        _cube = np.empty(shape=(shape[0], y_new, x_new))
        for idx, layer in enumerate(self.cube):
            _cube[idx] = cv2.resize(layer, (x_new, y_new), interpolation=mode)
        self.cube = _cube
        self.update_cube_shape()

    # todo: implemnt
    def shift_layers(self, num_layer: int) -> None:
        """
        Shift layers.

        :return: None
        """
        if num_layer > self.shape[0]:
            raise ValueError(f'`Num_layer` {num_layer} must me <= then the'
                             'layer deeps of the DataCube {self.shape[0]}')
        raise NotImplementedError('Sorry - Not Implemented')

    def set_wavelengths(self, wavelengths: np.array) -> None:
        """
        Set wavelength data.

        :return: None
        """
        if not isinstance(wavelengths, np.ndarray):
            try:
                # todo: better error handling
                if np.array(wavelengths).ndim == 1:
                    self.wavelengths = np.array(wavelengths)
                else:
                    raise AttributeError
            except AttributeError:
                raise AttributeError('Your wavelengths didnt match an'
                                     '1d np.array')

        else:
            if wavelengths.ndim == 1:
                self.wavelengths = wavelengths
            else:
                raise AttributeError('Your wavelengths didnt match an'
                                     '1d np.array')

    def set_cube(self, cube: np.array) -> None:
        """
        Set cube data.

        :return: None
        """
        if not isinstance(cube, np.ndarray):
            try:
                # todo: better error handling
                cube = np.array(cube)
            except AttributeError:
                raise AttributeError('Your cube is not convertable to a'
                                     'np.array')
        if 3 <= cube.ndim <= 4:
            self.cube = cube
        elif cube.ndim == 2:
            self.cube = np.zeros(shape=(1, cube.shape[0], cube.shape[1]),
                                 dtype=cube.dtype)
            self.cube[0] = cube
            print(f'\033[93mYour cube got forced to {self.cube.shape}\033[0m')
        else:
            raise AttributeError('Cube Data is not ndim 2,3 or 4')
        self.update_cube_shape()

    def update_cube_shape(self) -> None:
        """
        Update cube shape.

        :return: None
        """
        self.shape = self.cube.shape

    def start_recording(self):
        """
        Start Recording.

        :return: None
        """
        self.record = True
        TrackExecutionMeta.start_recording()

    def stop_recording(self) -> None:
        """
        Stop Recording.

        :return: None
        """
        self.record = False
        TrackExecutionMeta.stop_recording()

    @staticmethod
    def save_template(filename) -> None:
        """
        Save template from executed functions.

        :return: None
        """
        if not filename.endswith('.pickle'):
            filename = filename + '.pickle'
        with open(filename, 'wb') as template_file:
            pickle.dump(TrackExecutionMeta.recorded_methods, template_file)

    def load_template(self, filenmae) -> None:
        """
        Load template and execute function.

        :return: None
        """
        with open(filenmae, 'rb') as template_file:
            template_data = pickle.load(template_file)
        self.execute_template(template_data)




# Generated by CodiumAI
import numpy as np

from wizard import DataCube

import pytest


class TestDataCube:

    #  DataCube can be initialized with cube, wavelengths, name, and notation.
    def test_initialize_with_parameters(self):
        cube = np.array([[[1, 2], [3, 4]], [[5, 6], [7, 8]]])
        wavelengths = np.array([400, 500])
        name = "Cube1"
        notation = "A"
    
        data_cube = DataCube(cube=cube, wavelengths=wavelengths, name=name, notation=notation)
    
        assert data_cube.cube is not None
        assert data_cube.wavelengths is not None
        assert data_cube.name == name
        assert data_cube.notation == notation

    #  DataCube can be concatenated with another DataCube using the '+' operator.
    def test_concatenate_datacubes(self):
        cube1 = np.array([[[1, 2], [3, 4]], [[5, 6], [7, 8]]])
        wavelengths1 = np.array([400, 500])
        name1 = "Cube1"
        notation1 = "A"
    
        cube2 = np.array([[[9, 10], [11, 12]], [[13, 14], [15, 16]]])
        wavelengths2 = np.array([600, 700])
        name2 = "Cube2"
        notation2 = "B"
    
        data_cube1 = DataCube(cube=cube1, wavelengths=wavelengths1, name=name1, notation=notation1)
        data_cube2 = DataCube(cube=cube2, wavelengths=wavelengths2, name=name2, notation=notation2)
    
        result = data_cube1 + data_cube2
    
        assert result.cube is not None
        assert result.wavelengths is not None
        assert result.name == name1
        assert result.notation == notation1

    #  DataCube's shape can be accessed.
    def test_access_shape(self):
        cube = np.array([[[1, 2], [3, 4]], [[5, 6], [7, 8]]])
        data_cube = DataCube(cube=cube)
    
        shape = data_cube.shape
    
        assert shape == (2, 2, 2)

    #  DataCube's length can be accessed.
    def test_access_length(self):
        cube = np.array([[[1, 2], [3, 4]], [[5, 6], [7, 8]]])
        data_cube = DataCube(cube=cube)
    
        length = len(data_cube)
    
        assert length == 2

    #  DataCube's layers can be accessed using the [] operator.
    def test_access_layers(self):
        cube = np.array([[[1, 2], [3, 4]], [[5, 6], [7, 8]]])
        data_cube = DataCube(cube=cube)
    
        layer = data_cube[0]
    
        assert np.array_equal(layer, np.array([[1, 2], [3, 4]]))

    #  DataCube can be initialized with no parameters.
    def test_initialize_with_no_parameters(self):
        data_cube = DataCube()
    
        assert data_cube.cube is None
        assert data_cube.wavelengths is None
        assert data_cube.name is None
        assert data_cube.notation is None

    #  DataCube can be initialized with a cube but no wavelengths.
    def test_initialize_with_cube_no_wavelengths(self):
        cube = np.array([[[1, 2], [3, 4]], [[5, 6], [7, 8]]])
        data_cube = DataCube(cube=cube)
    
        assert data_cube.cube is not None
        assert data_cube.wavelengths is not None

    #  DataCube can be concatenated with another DataCube with different x and y shapes.
    def test_concatenate_datacubes_different_shapes(self):
        cube1 = np.array([[[1, 2], [3, 4]], [[5, 6], [7, 8]]])
        wavelengths1 = np.array([400, 500])
        name1 = "Cube1"
        notation1 = "A"
    
        cube2 = np.array([[[9, 10, 11], [12, 13, 14]], [[15, 16, 17], [18, 19, 20]]])
        wavelengths2 = np.array([600, 700])
        name2 = "Cube2"
        notation2 = "B"
    
        data_cube1 = DataCube(cube=cube1, wavelengths=wavelengths1, name=name1, notation=notation1)
        data_cube2 = DataCube(cube=cube2, wavelengths=wavelengths2, name=name2, notation=notation2)
    
        with pytest.raises(ValueError):
            result = data_cube1 + data_cube2

    # Initializes a DataCube instance with all parameters.
    def test_all_parameters(self):
        cube = np.zeros((10, 100, 100))
        wavelengths = [400, 410, 420, 430, 440, 450, 460, 470, 480, 490]
        name = "Example DataCube"
        notation = "wavelengths"
        record = True

        data_cube = DataCube(cube=cube, wavelengths=wavelengths, name=name, notation=notation, record=record)

        assert data_cube.name == name
        assert data_cube.shape == cube.shape
        assert data_cube.dim is None
        assert np.array_equal(data_cube.wavelengths, np.array(wavelengths))
        assert np.array_equal(data_cube.cube, cube)
        assert data_cube.notation == notation
        assert data_cube.record is True

    # Add a datacube and an int and check if the ValueError gets raised.
    def test_add_with_int_raises_valueerror(self):
        cube1 = DataCube(cube=np.ones((3, 2, 2)))
        cube2 = 5
        with pytest.raises(ValueError):
            cube1 + cube2

    # Adding two DataCubes with same shape and one of them has empty wavelengths should return a new DataCube with
    # concatenated cube and None wavelengths, and a warning message
    def test_same_shape_one_empty_wavelengths(self):
        dc1 = DataCube(cube=np.ones((10, 100, 100)), wavelengths=np.arange(10), name="dc1")
        dc2 = DataCube(cube=np.ones((10, 100, 100)), wavelengths=None, name="dc2")
        dc2.wavelengths = None

        with pytest.warns(UserWarning):
            result = dc1 + dc2

        assert isinstance(result, DataCube)
        assert result.cube.shape == (20, 100, 100)
        assert result.cube.shape[0] == len(result.wavelengths)

    #  Set a valid 3D numpy array to the cube attribute
    def test_set_valid_3d_array(self):
        cube_data = np.array([[[1, 2, 3], [4, 5, 6], [7, 8, 9]]])
        dc = DataCube()
        dc.set_cube(cube_data)
        assert np.array_equal(dc.cube, cube_data)

    #  Set a valid 4D numpy array to the cube attribute
    def test_set_valid_4d_array(self):
        cube_data = np.array([[[[1, 2, 3], [4, 5, 6], [7, 8, 9]]]])
        dc = DataCube()
        dc.set_cube(cube_data)
        assert np.array_equal(dc.cube, cube_data)

    #  Set a valid 2D numpy array to the cube attribute
    def test_set_valid_2d_array(self):
        cube_data = np.array([[1, 2, 3], [4, 5, 6], [7, 8, 9]])
        dc = DataCube()
        dc.set_cube(cube_data)
        expected_cube = np.zeros(shape=(1, cube_data.shape[0], cube_data.shape[1]), dtype=cube_data.dtype)
        expected_cube[0] = cube_data
        assert np.array_equal(dc.cube, expected_cube)

    #  Set a valid 3D numpy array to the cube attribute with the same shape as an existing cube
    def test_set_valid_3d_array_same_shape(self):
        existing_cube = np.array([[[1, 2, 3], [4, 5, 6], [7, 8, 9]]])
        dc = DataCube(cube=existing_cube)
        cube_data = np.array([[[10, 11, 12], [13, 14, 15], [16, 17, 18]]])
        dc.set_cube(cube_data)
        assert np.array_equal(dc.cube, cube_data)

    #  Set a valid 4D numpy array to the cube attribute with the same shape as an existing cube
    def test_set_valid_4d_array_same_shape(self):
        existing_cube = np.array([[[[1, 2, 3], [4, 5, 6], [7, 8, 9]]]])
        dc = DataCube(cube=existing_cube)
        cube_data = np.array([[[[10, 11, 12], [13, 14, 15], [16, 17, 18]]]])
        dc.set_cube(cube_data)
        assert np.array_equal(dc.cube, cube_data)

    #  Set an empty numpy array to the cube attribute
    def test_set_empty_array(self):
        cube_data = np.array([])
        dc = DataCube()
        with pytest.raises(AttributeError):
            dc.set_cube(cube_data)
        assert np.array_equal(dc.cube, None)

    #  Set a list to the cube attribute
    def test_set_list(self):
        cube_data = [[[1, 2, 3], [4, 5, 6], [7, 8, 9]]]
        dc = DataCube()
        dc.set_cube(cube_data)
        assert np.array_equal(dc.cube, np.array(cube_data))

    # test __iter__ and __next__
    def test_returns_next_item_and_wavelength(self):
        dc = DataCube(cube=np.random.rand(10, 100, 100), wavelengths=np.arange(10))
        iterator = iter(dc)
        next_value = next(iterator)
        assert isinstance(next_value, tuple)
        assert len(next_value) == 2
        assert isinstance(next_value[0], np.ndarray)

    # Should return a string with the name and shape of the DataCube
    def test_return_string_with_name_and_shape(self):
        dc = DataCube(cube=np.ones((3, 4, 5)), wavelengths=[400, 500, 600], name="Example")
        expected_output = ("Name: Example\n"
                           "Shape: (3, 4, 5)\n"
                           "Wavelengths\n"
                           "Num: 3\n"
                           "From: 400\n"
                           "To: 600\n"
                           "Cube:\n"
                           "[[[1. 1. 1. 1. 1.]\n  [1. 1. 1. 1. 1.]\n  [1. 1. 1. 1. 1.]\n  [1. 1. 1. 1. 1.]]\n\n"
                           " [[1. 1. 1. 1. 1.]\n  [1. 1. 1. 1. 1.]\n  [1. 1. 1. 1. 1.]\n  [1. 1. 1. 1. 1.]]\n\n"
                           " [[1. 1. 1. 1. 1.]\n  [1. 1. 1. 1. 1.]\n  [1. 1. 1. 1. 1.]\n  [1. 1. 1. 1. 1.]]]\n")

        assert dc.__str__() == expected_output

    # Resizing a cube with valid x_new and y_new values and linear interpolation
    def test_valid_resize_linear_interpolation(self):
        cube = np.random.rand(10, 100, 100)
        dc = DataCube(cube=cube)
        dc.resize(50, 50, interpolation='linear')
        assert dc.cube.shape == (10, 50, 50)

    # Resizing a cube with valid x_new and y_new values and nearest interpolation
    def test_valid_resize_nearest_interpolation(self):
        cube = np.random.rand(10, 100, 100)
        dc = DataCube(cube=cube)
        dc.resize(50, 50, interpolation='nearest')
        assert dc.cube.shape == (10, 50, 50)

    # Resizing a cube with x_new and y_new values
    def test_resize_to_up_value(self):
        cube = np.random.rand(10, 100, 100)
        dc = DataCube(cube=cube)
        dc.resize(200, 200, interpolation='linear')
        assert dc.cube.shape == (10, 200, 200)

    #  Method sets wavelength data when given a valid 1D numpy array.
    def test_valid_wavelengths_array(self):
        wavelengths = np.array([400, 450, 500, 550, 600])
        dc = DataCube()
        dc.set_wavelengths(wavelengths)
        assert np.array_equal(dc.wavelengths, wavelengths)

    #  Method sets wavelength data when given an empty 1D numpy array.
    def test_empty_wavelengths_array(self):
        wavelengths = np.array([])
        dc = DataCube()
        dc.set_wavelengths(wavelengths)
        assert np.array_equal(dc.wavelengths, wavelengths)

    #  Method raises AttributeError when given a non-convertible object.
    def test_non_convertible_object(self):
        wavelengths = "invalid"
        dc = DataCube()
        with pytest.raises(AttributeError):
            dc.set_wavelengths(wavelengths)

    #  Method raises AttributeError when given a numpy array with ndim != 1.
    def test_ndim_not_1(self):
        wavelengths = np.array([[400, 450, 500], [550, 600, 650]])
        dc = DataCube()
        with pytest.raises(AttributeError):
            dc.set_wavelengths(wavelengths)

    #  Method raises AttributeError when given a numpy array with ndim == 0.
    def test_ndim_0(self):
        wavelengths = np.array(400)
        dc = DataCube()
        with pytest.raises(AttributeError):
            dc.set_wavelengths(wavelengths)



# Generated by CodiumAI

from wizard._utils.data_loader import get_files_by_extension, make_path_absolute, to_cube, read

from wizard._core.datacube import DataCube

import pytest

import numpy as np


class TestGetFilesByExtension:

    #  Returns a sorted list of filenames with the given extension from a valid directory.
    def test_valid_directory(self):
        # Arrange
        path = '/path/to/directory'
        extension = '.csv'

        # Act
        result = get_files_by_extension(path, extension)

        # Assert
        assert isinstance(result, list)
        assert all(isinstance(filename, str) for filename in result)
        assert all(filename.endswith(extension) for filename in result)
        assert sorted(result) == result

    #  Returns an empty list if the extension is not provided.
    def test_no_extension_provided(self):
        # Arrange
        path = '/path/to/directory'
        extension = ''

        # Act
        result = get_files_by_extension(path, extension)

        # Assert
        assert isinstance(result, list)
        assert len(result) == 0

    #  Accepts a valid directory path with a trailing slash.
    def test_directory_with_trailing_slash(self):
        # Arrange
        path = '/path/to/directory/'
        extension = '.csv'

        # Act
        result = get_files_by_extension(path, extension)

        # Assert
        assert isinstance(result, list)
        assert all(isinstance(filename, str) for filename in result)
        assert all(filename.endswith(extension) for filename in result)
        assert sorted(result) == result


class TestMakePathAbsolute:

    #  The function receives a valid absolute path as input and returns the same path.
    def test_valid_absolute_path(self):
        # Arrange
        path = "/home/user/file.txt"

        # Act
        result = make_path_absolute(path)

        # Assert
        assert result == path

    #  The function receives an empty string as input and raises a ValueError.
    def test_empty_string_input(self):
        # Arrange
        path = ""

        # Act and Assert
        with pytest.raises(ValueError):
            make_path_absolute(path)

    #  The function receives a non-string input and raises a ValueError.
    def test_non_string_input(self):
        # Arrange
        path = 123

        # Act and Assert
        with pytest.raises(ValueError):
            make_path_absolute(path)

class TestToCube:

    #  Should transform a 2d numpy array to a 3d numpy array with the correct shape
    def test_transform_2d_to_3d(self):
        # Arrange
        data = np.array(
            [
            [
                [1, 2, 3],
                [4, 5, 6],
                [1, 2, 3]
            ],
            [
                [4, 5, 6],
                [1, 2, 3],
                [4, 5, 6]
            ],
            [
                [1, 2, 3],
                [4, 5, 6],
                [1, 2, 3]
                ],[
                [4, 5, 6],
                [1, 2, 3],
                [4, 5, 6]
            ]
        ]
        )
        len_x = 3
        len_y = 3

        # Act
        result = to_cube(data, len_x, len_y)

        # Assert
        assert result.shape == (4, 3, 3)


"""
class TestReadFSM:

    def test_load(self):

        path = '../resources/samples/20232509_SE.fsm'

        # spec = specread(path)
        dc = read(path)

        assert dc.shape == (408, 381, 729)
"""


import numpy as np
# from wizard import DataCube
from joblib import Parallel, delayed


def calculate_modified_z_score(cube):
    """
    Calculate the modified z-score of a data cube by computing the difference in intensity
    along the first axis.
    
    Parameters
    ----------
    cube : numpy.ndarray
        The input data cube of shape (Z, Y, X), where Z is the number of slices,
        Y is the height, and X is the width.
    
    Returns
    -------
    numpy.ndarray
        The modified z-score, which is the difference in intensity along the first axis.
    
    Example
    --------
    >>> cube = np.random.rand(10, 20, 30)  # Example data cube of shape (10, 20, 30)
    >>> modified_z_score = calculate_modified_z_score(cube)
    >>> print(modified_z_score.shape)  # Output: (20, 30)
    """
    delta_intensity = np.diff(cube, axis=0)
    return delta_intensity


def process_slice(spec_out_flat, spikes_flat, idx, window):
    """
    Process a single slice of the data cube to remove spikes by replacing them with the mean
    of the neighboring values within a given window.
    
    Parameters
    ----------
    spec_out_flat : numpy.ndarray
        Flattened output spectrum.
    spikes_flat : numpy.ndarray
        Flattened array indicating the presence of spikes.
    idx : int
        Index of the current slice to process.
    window : int
        Size of the window used to calculate the mean of neighboring values.
    
    Returns
    -------
    tuple
        A tuple containing the index of the processed slice and the modified slice.
    """
    
    w_h = int(window / 2)
    spike = spikes_flat[idx]
    tmp = np.copy(spec_out_flat[idx])

    for spk_idx in np.where(spike)[0]:
        window_min = max(0, spk_idx - w_h)
        window_max = min(len(tmp), spk_idx + w_h + 1)

        if window_min == spk_idx:
            window_data = tmp[spk_idx + 1:window_max]
        elif window_max == spk_idx + 1:
            window_data = tmp[window_min:spk_idx]
        else:
            window_data = np.concatenate((tmp[window_min:spk_idx], tmp[spk_idx + 1:window_max]))

        if len(window_data) > 0:
            tmp[spk_idx] = np.mean(window_data)
        else:
            tmp[spk_idx] = tmp[spk_idx]

    return idx, tmp


# @track_execution_time
def remove_spikes(dc, threshold: int = 6500, window: int = 3):
    """
    Remove cosmic spikes from a data cube using a specified threshold and window size.

    Parameters
    ----------
    dc : DataCube
        The input data cube from which spikes are to be removed.
    threshold : int, optional
        The threshold value for detecting spikes based on the modified z-score.
    window : int, optional
        The size of the window used to calculate the mean of neighboring values
        when replacing spikes.
    
    Returns
    -------
    numpy.ndarray
        The data cube with spikes removed.
    """
        
    z_spectrum = calculate_modified_z_score(dc.cube)
    spikes = abs(z_spectrum) > threshold
    cube_out = dc.cube.copy()

    spikes_flat = spikes.reshape(dc.cube.shape[0] - 1, -1)
    spec_out_flat = cube_out.reshape(cube_out.shape[0], -1)

    results = Parallel(n_jobs=-1)(
        delayed(process_slice)(spec_out_flat, spikes_flat, idx, window) for idx in range(spikes_flat.shape[0]))

    for idx, tmp in results:
        spec_out_flat[idx] = tmp

    spec_out_flat.reshape(cube_out.shape)
    return cube_out



import glob
import os

import re
import pathlib
import numpy as np
import pandas as pd

from tqdm import tqdm

from .fsm_reader import _read_fsm

from nptdms import TdmsFile
    
from matplotlib import pyplot as plt

from concurrent.futures import ThreadPoolExecutor

from .._core.datacube import DataCube

from wizard._utils.decorators import check_path, add_method


def get_files_by_extension(path: str, extension: str) -> list:
    """
    Return a sorted list of filenames of a given extension from a directory.

    :param path: path of a directory
    :type path: str
    :param extension: extension, for example `.csv`
    :rtype extension: str
    :return: list with filenames
    :rtype: list
    """
    # check if extension exists
    if not extension:
        return []

    # check if path is valid
    if not os.path.isdir(path):
        return []

    # check if extension doesn't start with `.`
    if not extension.startswith('.'):
        extension = '.' + extension

    return sorted(glob.glob(os.path.join(path, '*' + extension.lower())))


def make_path_absolute(path: str) -> str:
    """
    Check if the path is absolute. If not, convert it to an absolute path.

    :param path: Path to the file or directory
    :return: Absolute path to the file or directory
    :rtype: str
    """
    if isinstance(path, str):
        if not os.path.isabs(path):
            raise ValueError(f"Invalid path: {path}")
        return os.path.abspath(path).lower()
    else:
        raise ValueError("Input path must be a string.")


def to_cube(data, len_x, len_y) -> np.array:
    """
    Transform a 2d numpy array to a dc like array.

    data stored in Fortran-like index ordering, so using f-order instate c

    :param data: list of data in fortran order
    :param len_x: len dc x / pixel size x
    :param len_y: len dc y / pixel size y
    :return: transformed dc
    :rtype: np.array
    """
    len_v = data.shape[0]
    return data.reshape(len_v, len_x, len_y, order='F')


@check_path
@add_method(DataCube)
def read(path: str, datatype: str = 'auto', **kwargs) -> DataCube:
    """
    Read functions for importing data from different file types.

    :param path: data path to file
    :param datatype: data type of the file
    :param kwargs: additional keyword arguments
    :return: DataCube object
    """
    _datacube = None

    if datatype == 'auto':
        suffix = pathlib.Path(path).suffix
    else:
        suffix = datatype

    if suffix == '.csv':
        _datacube = __read_csv__(path, **kwargs)
    elif suffix == '.xlsx':
        _datacube = __read_xlsx__(path, **kwargs)
    elif suffix == '.fsm':
        _datacube = __read_fsm__(path, **kwargs)
    elif suffix == '.tdms':
        sample, dark, wave = read_tdms(path, **kwargs)
        _datacube = DataCube(sample, wavelengths=wave)
    elif suffix == '.jpg':
        _datacube = image_to_dc(path, **kwargs)
    elif suffix == '.npy':
        _datacube = load_npy(path, **kwargs)
    elif os.listdir(path):
        _datacube = images_from_folder_to_dc(path, **kwargs)
    else:
        raise NotImplementedError(f'No loader for {suffix}, '
                                  f'please parse your data manually.')

    return _datacube


def load_npy(path) -> DataCube:
    """
    Load numpy data.

    :param path:
    :return:
    """
    data = np.load(path)
    data = np.transpose(data, (2, 0, 1))

    # put data in DataCube and return
    return DataCube(data)


def load(data) -> DataCube:
    """
    Function for loading existing data into a DataCube.

    This function initializes a DataCube object from the provided data. It supports both numpy arrays and lists as input data formats.

    :param data: The data to be loaded into the DataCube. It can be either a numpy array or a list.
    :type data: numpy.ndarray or list
    :returns: A DataCube object initialized with the provided data, or None if the data format is unsupported.
    :rtype: DataCube or None
    """

    _datacube = None

    if isinstance(data, np.ndarray):
        _datacube = DataCube(cube=data)
    elif isinstance(data, list):
        _datacube = DataCube(cube=data)

    return _datacube


def filter_image_files(files):
    """
    Filters a list of filenames, returning only those that have image file extensions.

    The function checks for the following image file extensions (case-insensitive):
    - .jpg
    - .jpeg
    - .png
    - .gif
    - .bmp
    - .tiff

    :param files: A list of filenames to be filtered for image file extensions.
    :type files: list[str]
    :returns: A list of filenames that have image file extensions.
    :rtype: list[str]

    :Example:

    >>> files = ["image.jpg", "document.pdf", "photo.png", "archive.zip"]
    >>> image_files = filter_image_files(files)
    >>> print(image_files)  # Output: ['image.jpg', 'photo.png']
    """

    image_extensions = {".jpg", ".jpeg", ".png", ".gif", ".bmp", ".tiff"}
    return [file for file in files if any(file.lower().endswith(ext) for ext in image_extensions)]


@check_path
def images_from_folder_to_dc(path: str, **kwargs) -> DataCube:
    """
    Load a folder of images into a DataCube.

    #todo: add exclude parameter to avoid some files or to an
    include_only parameter

    :param path:
    :return:
    """
    _files = [os.path.join(path, f) for f in os.listdir(path)]

    _files_filtert = filter_image_files(_files)

    _dc = image_to_dc(_files_filtert, **kwargs)

    # put data in DataCube and return
    return _dc


def load_image(path):
    """
    Load an image from a specified file path.

    Parameters
    ----------
    path : str
        The file path to the image to be loaded.

    Returns
    -------
    ndarray
        The image read from the file, represented as a NumPy array.

    Examples
    --------
    >>> img = load_image('path/to/image.png')
    >>> plt.imshow(img)
    >>> plt.show()
    """
    return plt.imread(path)


def image_to_dc(path: str | list, **kwargs) -> DataCube:
    """
    Load image(s) into a DataCube.

    This function loads one or more images into a DataCube. It supports both a single image file path or a list of image file paths. The images are processed based on the specified type, which determines the transpose operation applied to the data.

    :param path: Path to an image file or a list of image file paths. If a list is provided, images are loaded concurrently.
    :type path: str or list[str]
    :param kwargs: Optional keyword arguments.
        - type: Specifies the transpose operation to apply to the data. Can be 'default' (default behavior) or 'pushbroom' (for pushbroom images).
        - Other keyword arguments may be accepted depending on the implementation of `load_image`.

    :returns: A DataCube object containing the image data.
    :rtype: DataCube

    :raises TypeError: If `path` is neither a string nor a list of strings.
    """

    type = kwargs.get('type', 'default')
    name = kwargs.get('name', None)

    if isinstance(path, str):
        img = load_image(path)
        data = np.transpose(np.array(img), (2, 0, 1))
        
    elif isinstance(path, list):

        def process_image(idx_file):
            idx, file = idx_file
            _img = load_image(file)
            return _img

        with ThreadPoolExecutor() as executor:
            results = list(executor.map(process_image, enumerate(path)))

        data = np.array(results)

        if type == 'pushbroom':
            data = np.transpose(data, (1, 0, 2))
        else:
            data = np.transpose(data, (2, 0, 1))

    else:
        raise TypeError('Path must be string to a file or a list of files')
    
    return DataCube(data, name=name)


# todo
def __read_csv__(path: str) -> DataCube:
    """
    Read csv file.

    :param path: path to csv file
    :return:
    """
    raise NotImplementedError('Not Implemented yet')
    _datacube = DataCube(cube=None)

    return _datacube


# todo
def __read_xlsx__(path: str) -> DataCube:
    """
    Read xlsx file.

    :param path: path to xslx file
    :return:
    """
    raise NotImplementedError('Not Implemented yet')
    _datacube = DataCube(cube=None)

    return _datacube


def __read_fsm__(path: str) -> DataCube:
    """
    Read function for fsm-files from perkin elmer. Tested with FTIR-Data.

    ToDo: missing error handling if `path` is wrong.

    :param path: str, path to file
    :return:
    """
    # load data - load spectra from path
    # fsm_spectra = specread(path)
    fsm_spectra, fsm_wave, fsm_meta = _read_fsm(path)

    # load data - load x&y lens
    fsm_len_x = fsm_meta['n_x']
    fsm_len_y = fsm_meta['n_y']
    # fsm_len_y = fsm_meta['n_z']

    # load data - start and stop from the wavelength
    # fsm_wave_start = fsm_meta['z_start']
    # fsm_wave_end = fsm_meta['z_end']

    # load data - wavelength
    fsm_wave = fsm_wave.astype('int')

    # load data - data
    # fsm_data = fsm_spectra.amplitudes

    # ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    # convert 2d df in 3d np.array
    # ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

    fsm_data_cube = to_cube(data=fsm_spectra.T, len_x=fsm_len_x, len_y=fsm_len_y)

    # may return with transposed date
    _datacube = DataCube(fsm_data_cube, wavelengths=fsm_wave,
                         name='.fsm', notation='cm-1')
    return _datacube


def read_tdms(path: str = None) -> DataCube:
    """
    Read function for tdms file.

    The functions reads and pareses the in `tdms_file` defined file and
    returns the data in a defined way. The return value can be `np` or df`.

    :param path: Path in string format with reference to the file to
                 be read.
    :return: datacube
    :rtype: wizard.DataCube
    """
    # type for automatic detection
    data_type = ''
    wave_col = 0
    len_col = 0

    # get values
    file = TdmsFile(path)

    # build df
    tdms_df = file.as_dataframe()

    # copy cols
    col = tdms_df.columns
    col_new = []
    col_sample = []
    col_raw = []

    # sort by dark and normal
    for i in col:

        i = i.replace(' ', '').replace('\'', '')

        if i.find('RAW') >= 1:
            col_raw.append(i)
        elif (i.find('DarkCurrent') >= 1 or i.find('cm') >= 1 or i.find('nm') >= 1):
            pass
        else:
            col_sample.append(i)

        col_new.append(i)

    # rename cols
    tdms_df.columns = col_new

    if any("RAMAN" in s for s in col_new):
        data_type = 'raman'
        wave_col = 1
        len_col = 4

    if any("NIR" in s for s in col_new) or any("KNIR" in s for s in col_new):
        data_type = 'nir'
        wave_col = 1
        len_col = 3

    if any("VIS" in s for s in col_new) or any("KVIS" in s for s in col_new):
        data_type = 'vis'
        wave_col = 1
        len_col = 3

    # get wave length
    # wave_col: - 1 raman for cm-¹; -2 for nm
    wave = np.array(tdms_df[tdms_df.columns[-wave_col]])

    # parse length information
    # len_col: -3 for other,  -4 for raman
    len_xy = re.findall(r'\d+', col_new[-len_col])
    # len_v = wave.shape[0]
    len_x = int(len_xy[0]) + 1
    len_y = int(len_xy[1]) + 1

    # set index
    tdms_df = tdms_df.set_index(tdms_df.columns[-2])

    # cops into new df
    tdms_sample_df = tdms_df[col_sample].copy()
    # tdms_raw_df = tdms_df[col_raw].copy()

    # clean up
    del tdms_df

    tdms_sample = np.array(tdms_sample_df)
    # tdms_raw = np.array(tdms_raw_df)

    tdms_sample_cube = to_cube(data=tdms_sample, len_x=len_x, len_y=len_y)
    # tdms_raw_cube = to_cube(data=tdms_raw, len_x=len_x, len_y=len_y)

    wave = wave.astype('int')

    return DataCube(
        cube=tdms_sample_cube,
        wavelengths=wave,
        name=data_type
    )


def write_xlsx(datacube: np.array, wavelenghts: np.array, filename: str):
    """
    Write out a .xlsx file.

    :param datacube:
    :param wavelenghts:
    :param filename:
    :return:
    """
    shape = datacube.shape

    df = pd.DataFrame()

    cols = []

    for i in wavelenghts:
        cols.append(str(i))

    idx = []

    for y in range(shape[1]):
        for x in range(shape[0]):
            spec_ = datacube[x, y, :]

            df_tmp = pd.DataFrame(spec_).T

            df = df.append(df_tmp)

            idx.append(f'x:{x}; y:{y}')

    df.columns = cols

    df.insert(0, column='Point', value=idx)

    df = df.set_index('Point')

    df.to_excel(f'{filename}.xlsx')


def merge_cubes(cube1: DataCube, cube2: DataCube) -> DataCube:
    """
    Merge to datacubes to a new one.

    :param cube1:
    :param cube2:
    :return:
    """
    c1 = cube1.cube
    c2 = cube2.cube
    if c1.shape[:2] == c2.shape[:2]:
        c3 = np.concatenate(c1, c2)
    else:
        c3 = None
        raise NotImplementedError('Sorry - '
                                  'This function is not implemented yet.'
                                  'At the moment you just can merge cubes'
                                  ' with the same size x,y.')
    return DataCube(c3)


def merge_waves(wave1: list, wave2: list) -> list:
    """
    Merge two wave lists.

    todo: better merge algorithms

    :param wave1: first list with waves
    :param wave2: second list with waves
    :return: merged waves
    :rtype: list
    """
    if common_members(wave1, wave2):
        raise NotImplementedError('Sorry - your wavelengths are overlapping,'
                                  ' we working on a solution')

    return wave1 + wave2


def common_members(a: list, b: list) -> set:
    """
    Check for comon members between two lists.

    :param a: list a
    :param b: list to compare to a
    :return: return a set of common members
    :rtype: set
    """
    a_set = set(a)
    b_set = set(b)
    return a_set & b_set



import yaml


if __name__ == '__main__':

    import os
    file = os.path.join(os.getcwd(), '../../resources', 'hsi_workflow.yml')
    with open(file, 'r') as stream:
        try:
            print(yaml.safe_load(stream))
        except yaml.YAMLError as exc:
            print(exc)



"""DataCube Tracker to keep record of used methods and functions."""
exculted = ['stop_recording', 'save_template']


class TrackExecutionMeta(type):
    """
    DataCube tracker.

    Keep track of used functions and methods.
    """

    recording = False
    recorded_methods = []

    def __new__(cls, name, bases, dct):
        """
        Magic Method new.

        :return:
        """
        for key, value in dct.items():
            if callable(value) and key != 'execute_template':
                dct[key] = cls.record_method(value)
        return super().__new__(cls, name, bases, dct)

    @staticmethod
    def record_method(func):
        """
        Record Decorator.

        :return: func
        """
        def wrapper(*args, **kwargs):
            if TrackExecutionMeta.recording:
                print(func.__name__)
                if func.__name__ not in exculted:
                    print('execute!')
                    TrackExecutionMeta.recorded_methods.append(
                        (func.__name__, args, kwargs))
            return func(*args, **kwargs)
        return wrapper

    @staticmethod
    def start_recording():
        """
        Start tracker.

        :return: None
        """
        TrackExecutionMeta.recording = True
        TrackExecutionMeta.recorded_methods = []

    @staticmethod
    def stop_recording() -> None:
        """
        Stop tracker.

        :return: None
        """
        TrackExecutionMeta.recording = False
